/*
  Program:      Mine Mayhem
  File:         board.cc
  Date: 	July, 1996 to February,   1997,	version 1.00.
		26 October to 3 December,	version 2.00.
		8 December to 4 January,  2001, version 2.10.
		15 January,			version 2.11.

  Everything to do with the board.

  Freeware, Copyright 2001 Jason Hood.

  You are free to use this code, or a portion thereof, as long as an
  appropriate acknowledgement is made.

  v2.00: timing now uses uclock(), with option to use GetTicks();
	 removed all minefield code;
	 added "shadow box" around the counters and timer;
	 pause is tested in the event section.
  v2.10: save the board as a bitmap;
	 replaced DisplayMines() & DisplaySpaces() with DisplayProgress(),
	  which also uses the millisecond display for percentage completed;
	 added tracking and auto-clear options.
*/

#include "board.h"
#include "minefld.h"
#include "faces.h"
#include <tws/system.h>
#include <stdio.h>
#include <unistd.h>

extern bool use_normal_timing;
extern int ClearButton; 	// options.h


#define ID_FACE 1000

DEFINE_RESPONSE_TABLE(FaceButton, BasicButton)
  E_RBUTTONDOWN
END_RESPONSE_TABLE


FaceButton::FaceButton(Window* parent, ControlID ID, int x, int y, int icon) :
  IconButton(parent, ID, x, y, x+FACEICONW-1, y+FACEICONH-1, FaceIcon[icon]),
  curicon(icon)
{
}

void FaceButton::ChangeIcon(int NewIcon)
{
  IconButton::ChangeIcon(FaceIcon[NewIcon]);
  curicon = NewIcon;
}

// Change the icon and display it as well.
void FaceButton::Display(int Icon)
{
  ChangeIcon(Icon);
  Paint();
  RefreshWindow();
}

// Display a different icon when the button is pressed.
void FaceButton::PaintWindow(int x1, int y1, int x2, int y2)
{
  IconButton::ChangeIcon(FaceIcon[isDown ? FACEPUSHED : curicon]);
  IconButton::PaintWindow(x1, y1, x2, y2);
}

void FaceButton::RButtonDown(int, int, int)
{
  ws.QueueEvent(parent, event(W_NOTIFY, butID, 1003));
}


DEFINE_RESPONSE_TABLE(BoardWindow, Window)
  E_BUTTONUP(ID_FACE, cmNewGame)
  if (!isPaused)
  {
    E_LBUTTONDOWN
    E_RBUTTONDOWN
    E_MBUTTONDOWN
    E_LBUTTONUP
    E_RBUTTONUP
    E_MBUTTONUP
    E_MOUSEMOVE
    E_TIMER
  }
  E_BUTTONR(ID_FACE,	cmPause)
  E_KEYCODE(0x70, 0x19, cmPause);	// P
  E_KEYCODE(0x00, 0x86, cmDump);	// F12
END_RESPONSE_TABLE


/* ----------------------------   BoardWindow	-----------------------------

  Create the board window, big enough to correctly position the face button.

*/

BoardWindow::BoardWindow() :
  Window(NULL, NULL, (ws.GetDeskWidth() - (FACEICONW + 4*2 + FWIDTH*2))/2,
                     SysFont->height + FWIDTH*2 + 4 + 4,
                     (ws.GetDeskWidth() + (FACEICONW + 4*2 + FWIDTH*2))/2 - 1,
                     SysFont->height + FWIDTH*2 + 4 + 4 +
                      4 + DIGITH + 4 + FACEICONH + FWIDTH*2 - 1,
	 WA_VISABLE | WA_SAVEAREA | WA_BORDER),
  isPaused(false)
{
  Face = new FaceButton(this, ID_FACE, FWIDTH+4, FWIDTH+4+DIGITH+4, FACESMILEY);
}


BoardWindow::~BoardWindow()
{
  delete Face;
}


/* -------------------------------   NewGame   ------------------------------

  Start a new game with the given dimensions. If the dimensions are not the
  same as the previous game resize the window.

*/

void BoardWindow::NewGame(cCoord& dim, int mines, int spaces)
{
  ws.StopTimer(this);
  if (isPaused)
  {
    isPaused = false;
    DeleteBuffer(gridsave);
  }
  Face->Display(FACESMILEY);
  secs = mils = 0;
  if (!(olddim == dim))
  {
    olddim = dim;
    Resize(dim.x, dim.y);
  }
  DisplayTime(true);
  progress = true;
  DisplayProgress(mines, spaces, UPDATE_BOTH);
}


// Start a new game from the face button.
void BoardWindow::cmNewGame()
{
  field->NewGame(field->level);
}


/* -------------------------------   Resize   -------------------------------

  Set the board's size and position according to the dimensions.

*/

void BoardWindow::Resize(int width, int height)
{
  //	    grid     frame   margin outline
  int wid = width + FWIDTH*2 + 4*2 + 2,

  //	     <------------ menu ---------->  margin
      ypos = SysFont->height + FWIDTH*2 + 4 + 4;

  Face->ShiftX((wid-FACEICONW)/2 - Face->x1);   // Re-centre the face button
  Window::Resize((ws.GetDeskWidth()-wid)/2, ypos, (ws.GetDeskWidth()+wid)/2-1,
  //   margin time  margin  button   margin outline  grid     frame
  ypos + 4 + DIGITH + 4 + FACEICONH + 4*2 +    2 +  height + FWIDTH*2 - 1);
}


/* ----------------------------   PaintWindow	-----------------------------

  Draw the outline around the grid, the frames around the counts and time, and
  display the dimensions and difficulty.
  Calculate some display offsets.

*/

void BoardWindow::PaintWindow(int x1, int y1, int x2, int y2)
{
  Window::PaintWindow(x1, y1, x2, y2);

  ViewBuffer* buf = GetSubBuffer(x1, y1, x2, y2);

  // Top-left co-ordinate of the grid itself
  ox	  = cx1 + 4 + 1;
  oy	  = cy1 + 4 + DIGITH + 4 + FACEICONH + 4 + 1;
  spacesx = cx2 - 5 - 3*DIGITW; 	// Offset for space count
  timex   = Face->x2 - 3*DIGITW - 1;	// Offset for time, above the face
					//  button, which is already centred

  // Draw the "shadow box" (Coyote's term) around the counts and time.
  // This does actually look better than version 1.
  #define F4 FWIDTH+4
  #define DW 3*DIGITW
  #define DH DIGITH
  InvFrame(buf, F4-1, F4+DH+3, F4+DW, F4+DH+4+DH);
  InvFrame(buf, FWIDTH+spacesx-1, F4+DH+3, FWIDTH+spacesx+DW, F4+DH+4+DH);
  InvFrame(buf, FWIDTH+timex-1, F4-1, FWIDTH+timex+DW, F4+DH);
  InvFrame(buf, FWIDTH+timex+DW, F4+DH-SMALLDIGITH-1,
		FWIDTH+timex+DW+3*SMALLDIGITW, F4+DH);
  Pixel(buf, FWIDTH+timex+DW, F4+DH, 17);
  #undef F4
  #undef DW
  #undef DH

  field->Outline(buf, ox-1, oy-1, cx2-4, cy2-4);

  char num[8];
  // Display the dimensions in the top-left corner ...
  sprintf(num, "%d x %d", field->layout.dim.x, field->layout.dim.y);
  WriteText(buf, FWIDTH+4, FWIDTH+4, 0, SysFont, num);
  // ... and the difficulty in the top-right
  sprintf(num, "%.3g", field->layout.diff());
  WriteText(buf, cx2-4-StringWidth(num, SysFont), FWIDTH+4, 0, SysFont, num);

  DeleteBuffer(buf);
}


/* ----------------------------   StartTiming	-----------------------------

  Initialise the time calibration and start the timer.

*/

void BoardWindow::StartTiming()
{
  oldsecs   = 0;
  startmils = (use_normal_timing) ? GetTicks() : uclock();
  ws.StartTimer(this, 250);
}


/* -------------------------------   Timer   --------------------------------

  The timer routine, called every quarter-second. It reads the system time and
  updates the display if the seconds have changed.

*/

void BoardWindow::Timer()
{
  if (isFocus())
  {
    if (use_normal_timing) secs = (GetTicks() - startmils) / 1000;
    else		   secs = (uclock()   - startmils) / UCLOCKS_PER_SEC;
    if (secs != oldsecs)
    {
      oldsecs = secs;
      DisplayTime();
    }
  }
}


/* ----------------------------   DisplayTime   -----------------------------

  Display the time. If full is true then it displays the milliseconds, other-
  wise only the seconds are displayed.

*/

void BoardWindow::DisplayTime(bool full)
{
  Display(secs, timex, 4);
  if (full)
    Display(mils, timex+3*DIGITW, 4+DIGITH-SMALLDIGITH, SmallDigit, true);
}


/* ------------------------------   GameOver   ------------------------------

  Draw the appropriate face, display the final time (ie. including the milli-
  seconds) and return the time taken (in milliseconds).

*/

int BoardWindow::GameOver(bool dead)
{
  int score;
  if (use_normal_timing) score = GetTicks() - startmils;
  else
    score = (uclock() - startmils) / (UCLOCKS_PER_SEC / 1000);
  ws.StopTimer(this);
  secs = score / 1000; mils = score % 1000;
  DisplayTime(true);
  Face->Display((dead) ? FACEDEAD : FACECOOL);
  progress = dead;
  return score;
}


/* ---------------------------	 DisplayProgress   ---------------------------

  Display the number of mines marked and/or the number of spaces to uncover.
  If playing the clear game, use the millisecond time display to indicate the
  percentage completed.

*/

void BoardWindow::DisplayProgress(int mines, int spaces, int update)
{
  if (update & UPDATE_MINES)  Display(mines, 4, 4+DIGITH+4);
  if (update & UPDATE_SPACES) Display(spaces, spacesx, 4+DIGITH+4);

  if (Game == CLEAR && progress)
  {
    int pc, t;
    int m = field->layout.mines;
    if (Finish == ALL_MARKED)
    {
      pc = 100 * (m - mines) / m;
    }
    else
    {
      t = field->area(field->layout.dim) - m;
      pc = 100 * (t - spaces) / t;
      if (Finish != ALL_OPENED)
      {
	pc = (pc + (100 * (m - mines) / m)) / 2;
      }
    }
    Display(pc, timex+3*DIGITW, 4+DIGITH-SMALLDIGITH, SmallDigit);
  }
}


/* ------------------------------   Display   -------------------------------

  Display the number num at (x, y) (relative to the board window's client
  area).  It uses the bmp[] buffers for each digit (0 to 9, 10 for minus, 11
  for blank, all the same dimensions).  If zeros is true the number is padded
  with zeros, otherwise blanks.  num is assumed to be three or less characters
  (ie.  -99 .. 999). (It can be four, in which case only the first three are
  printed; more than that will probably crash.)

*/

void BoardWindow::Display(int num, int x, int y, ViewBuffer* bmp[], bool zeros)
{
  char str[5];
  int n, x1, j,
      wid = bmp[0]->width, hyt = bmp[0]->height;
  x += cx1; y += cy1;

  sprintf(str, (zeros) ? "%03d" : "%3d", num);
  for (x1 = x, j = 0; j < 3; x1 += wid, j++)
  {
	 if (str[j] == ' ') n = BLANK;
    else if (str[j] == '-') n = MINUS;
    else		    n = str[j] - '0';
    BitBlt(wnd, x1, y, bmp[n], 0, 0, wid-1, hyt-1);
  }
  RefreshWindow(absx+x, absy+y, absx+x1-1, absy+y+hyt-1);
}


/* ---------------------------	 DisplayTile   ----------------------------

  Display the tile on the board. If refresh is false then the tile
  is not refreshed onto the screen.

*/

void BoardWindow::DisplayTile(int x, int y, int w, int h,
			      ViewBuffer* tile, bool refresh)
{
  x += ox; y += oy;
  BitBltTransparent(wnd, x, y, tile, 0, 0, w, h);
  if (refresh)
    RefreshWindow(absx+x, absy+y, absx+x+w, absy+y+h);
}


/* ----------------------------   ?ButtonDown	-----------------------------

  Normalise the co-ordinate and pass it along to the field.

*/

static Tile* old_t = NULL;
static int   butt = 0;

void BoardWindow::LButtonDown(int x, int y, int)
{
  x -= absx+ox; y -= absy+oy;
  field->Convert(x, y, old_t);
  butt |= MBS_LEFT_BUTTON;
  field->LeftButton(x, y);
}

void BoardWindow::RButtonDown(int x, int y, int)
{
  x -= absx+ox; y -= absy+oy;
  field->Convert(x, y, old_t);
  butt |= MBS_RIGHT_BUTTON;
  field->RightButton(x, y);
}

void BoardWindow::MButtonDown(int x, int y, int)
{
  x -= absx+ox; y -= absy+oy;
  field->Convert(x, y, old_t);
  butt |= MBS_MIDDLE_BUTTON;
  field->MiddleButton(x, y);
}


/* -----------------------------   ?ButtonUp   ------------------------------

  These are called when the button is released, to remove the highlight.

*/

void BoardWindow::LButtonUp(int, int, int)
{
  if (ClearButton == CLRLEFT) field->UnHighLight();
  butt &= ~MBS_LEFT_BUTTON;
}

void BoardWindow::RButtonUp(int, int, int)
{
  if (ClearButton == CLRRIGHT) field->UnHighLight();
  butt &= ~MBS_RIGHT_BUTTON;
}

void BoardWindow::MButtonUp(int, int, int)
{
  if (ClearButton == CLRMIDDLE) field->UnHighLight();
  butt &= ~MBS_MIDDLE_BUTTON;
}


/* -----------------------------   MouseMove   ------------------------------

  If tracking is on, follow the mouse. Only track if one button is pressed.

*/

void BoardWindow::MouseMove(int x, int y, int, int but_stat)
{
  Tile* t;

  if (!isTracked || butt == 0) return;

  x -= absx+ox; y -= absy+oy;
  if (field->Convert(x, y, t) && t != old_t)
  {
    switch (but_stat)
    {
      case MBS_LEFT_BUTTON:
	if (ClearButton == CLRLEFT) field->UnHighLight();
	field->LeftButton(x, y);
      break;
      case MBS_RIGHT_BUTTON:
	if (ClearButton == CLRRIGHT) field->UnHighLight();
	field->RightButton(x, y);
      break;
      case MBS_MIDDLE_BUTTON:
	if (ClearButton == CLRMIDDLE) field->UnHighLight();
	field->MiddleButton(x, y);
      break;
    }
    old_t = t;
  }
}


/* ------------------------------   cmPause   -------------------------------

  Pause the game. Display a sleepy face and blank the board. Call again to
  unpause.

*/


void BoardWindow::cmPause(int)
{
  if (!field->GameinProgress()) return;

  Coord dim = field->GetBoardDimensions(field->layout.dim);
  int wid = dim.x, hyt = dim.y;

  if (isPaused)
  {
    Face->Display(FACESMILEY);
    isPaused = false;
    BitBlt(wnd, ox, oy, gridsave, 0, 0, wid-1, hyt-1);
    DeleteBuffer(gridsave);
    RefreshWindow(absx+ox, absy+oy, absx+ox+wid-1, absy+oy+hyt-1);
    startmils += ((use_normal_timing) ? GetTicks() : uclock()) - pausemils;
  }
  else
  {
    pausemils = (use_normal_timing) ? GetTicks() : uclock();
    Face->Display(FACESLEEPY);
    isPaused = true;
    gridsave = NewBuffer(wid, hyt);
    BitBlt(gridsave, 0, 0, wnd, ox, oy, ox+wid-1, oy+hyt-1);
    field->Pause();
    RefreshWindow(absx+ox, absy+oy, absx+ox+wid-1, absy+oy+hyt-1);
  }
}


/* -------------------------------   cmDump   --------------------------------

  Save the board as a bitmap. Complicated by the fact that TWS will only save
  24-bit colour bitmaps and there is no get pixel routine (I'll assume the
  config remains 8-bit).

*/

void BoardWindow::cmDump(int)
{
  // Create the first 20 system colours.
  unsigned char rgb[20*3];
  int r, g, b, c;
  sscanf(ws.GetConfigItem("system_colour"), "%d,%d,%d", &r, &g, &b);
  c = 3;
  rgb[0] = rgb[1] = rgb[2] = 0;
  for (int j = 1; j < 20; ++j)
  {
    rgb[c++] = r * j / 19;
    rgb[c++] = g * j / 19;
    rgb[c++] = b * j / 19;
  }

  // Translate the 8-bit buffer to a 24-bit buffer. BitBlt should do this,
  // but doesn't.
  ViewBuffer* buf = NewBuffer24(wnd->width, wnd->height);
  unsigned char* pos = (unsigned char*)wnd->base;
  for (int y = 0; y < wnd->height; ++y)
  {
    for (int x = 0; x < wnd->width; ++x)
    {
      c = *pos++;
      if (c < 20)
      {
	c *= 3;
	r = rgb[c];
	g = rgb[c+1];
	b = rgb[c+2];
      }
      else // if (c <= 235)	// Custom colours aren't used
      {
	c -= 20;
	r = c / 36 * 51; c %= 36;
	g = c / 6  * 51;
	b = c % 6  * 51;
      }
      RGBPixel(buf, x, y, r, g, b);
    }
  }

  // Find the first available filename: mine0001.bmp, mine0002.bmp, ...
  char name[13];
  for (int num = 1; ; ++num)
  {
    sprintf(name, "mine%04d.bmp", num);
    if (access(name, F_OK)) break;
  }
  FILE* dump = fopen(name, "wb");
  SaveBMP(buf, dump);
  fclose(dump);

  DeleteBuffer(buf);
}
