/*
  Program:      Mine Mayhem
  File: 	fieldhex.cc
  Date: 	26 October to 3 December, 1997, version 2.00.
		8 December to 4 January,  2001, version 2.10.
		15 January,			version 2.11.

  Everything to do with the hexagonal minefield.

  A lot of this code is based on Coyote's Hexagonal Minesweeper. I really
  thank him for providing me with his source.

  Freeware, Copyright 2001 Jason Hood.

  You are free to use this code, or a portion thereof, as long as an
  appropriate acknowledgement is made.


  The minefield looks like:	    x x x
				   x x x x
				  x x x x x
				   x x x x
				    x x x

  which is stored as:		  x x x
				  x x x x
				  x x x x x    ^
				    x x x x    | height
				      x x x    v

				      |---| width

  v2.10: removed PanLevel() function, use pan_wid variable instead.
*/


#include "minefld.h"
#include "options.h"
#include "tiles.h"
#include "board.h"


HexagonField::HexagonField() : Minefield()
{
  Beginner	 = Layout( 5,  5, 10);
  Intermediate	 = Layout(10, 10, 45);
  Expert	 = Layout(13, 13, 95);

  MinWidth	 =
  MinHeight	 = 5;
  totalcols	 = max_width / HEXAGONW;
  rows_allocated = (max_height - HEXAGOND) / HEXAGONR;
  MaxHeight	 = (rows_allocated+1) / 2;

  CreateMinefield(totalcols);
}


/* --------------------------------   Area   --------------------------------

  Area of a "hexagon". I derived this myself. Is there a simpler equation?

*/

int HexagonField::area(cCoord& dim)
{
  return (dim.y*dim.y + 2*dim.x*dim.y - dim.x - 2*dim.y + 1);
}


/* ----------------------------   GetMaxWidth	-----------------------------

  The maximum "width" of the hexagon is dependant on its height.  (The
  _overall_ maximum width doesn't change (cf. totalcols), but that's not the
  dimension I'm using.)

*/

int HexagonField::GetMaxWidth(int height)
{
  return (MaxWidth = totalcols - height + 1);
}


/* -------------------------   CrossStart and End   -------------------------

  Return the beginning and finishing positions for the cross game.

*/

Tile* HexagonField::CrossStart()
{
  return tiles[0];
}

Tile* HexagonField::CrossEnd()
{
  return tiles[tile_count-1];
}


/* ----------------------------   RandomLayout	 ----------------------------

  Calculate the dimensions to satisfy the difficulty. diff is assumed to be
  valid (ie. between 2 and 10). There will always exist a layout to satisfy
  the difficulty (proven by experimentation).

*/

void HexagonField::RandomLayout(int diff)
{
  int a;
  do
  {
    Random.dim.y = random() % (MaxHeight - MinHeight + 1) + MinHeight;
    Random.dim.x = random() % (GetMaxWidth(Random.dim.y) - MinWidth + 1) +
		     MinWidth;
    a = area(Random.dim);
  }
  while (a % diff != 0);
  Random.mines = a / diff;
}


/* -----------------------------   AdjustGame	-----------------------------

  When the game changes from CLEAR to CROSS, need to ensure that the top and
  bottom corners are free from mines.

*/

void HexagonField::AdjustGame()
{
  bool moving = true;
  while (moving)
  {
    moving = (MoveMine(tiles[0])			   ||
	      MoveMine(tiles[1])			   ||
	      MoveMine(tiles[layout.dim.x])		   ||
	      MoveMine(tiles[layout.dim.x+1])		   ||
	      MoveMine(tiles[tile_count-2 - layout.dim.x]) ||
	      MoveMine(tiles[tile_count-1 - layout.dim.x]) ||
	      MoveMine(tiles[tile_count-2])		   ||
	      MoveMine(tiles[tile_count-1]));
  }
}


/* -----------------------------   AdjustWrap	-----------------------------

  I can't think of a way of providing wrap-around for the hexagonal field.
  It doesn't really fold around itself. The best I can come up with is cutting
  off the bottom line, allowing the top line to fold in its place, creating a
  cylinder. The left and right edges still wouldn't wrap, though.

*/

void HexagonField::AdjustWrap()
{
}


/* ------------------------------   Convert   -------------------------------

  Convert the co-ordinate to a tile.
  Returns false if the co-ordinate is not on the grid, otherwise true.
  false is also returned if the hexagon is not next to an open hexagon in
  the cross game.

  This is not a perfect conversion, but it's good enough. (Blame Coyote. :) )

*/

bool HexagonField::Convert(int x, int y, Tile*& t)
{
  Coord b = GetBoardDimensions(layout.dim);
  y -= HEXAGOND/2;
  if (x < 0 || x >= b.x ||
      y < 0 || y >= b.y) return false;

  y /= HEXAGONR;
  x -= (HEXAGONW/2) * (layout.dim.y-1 - y);
  if (x < 0) return false;
  x /= HEXAGONW;

  if (y >= layout.dim.y * 2 - 1) return false;
  if (y < layout.dim.y)
  {
    if (x >= layout.dim.x + y) return false;
  }
  else if (x <= y - layout.dim.y || x >= layout.dim.x + layout.dim.y - 1)
    return false;

  t = &grid[y][x];
  return CrossValid(t);
}


/* -----------------------------   EveryTile   ------------------------------

  Create the list of hexagons, their x-co-ordinates and surrounding hexagons.

  If create is true, set the y-co-ordinate and allocate memory for the
  surrounding hexagons.

*/

void HexagonField::EveryTile(bool create)
{
  if (!create)
  {
    tile_count = area(layout.dim);
    int pos = 0;
    int x1, lx = layout.dim.x + layout.dim.y - 1;
    for (int y = 0; y < layout.dim.y * 2 - 1; y++)
    {
      x1 = (y < layout.dim.y) ? y + layout.dim.x : lx;
      for (int x = (y < layout.dim.y) ? 0 : y - (layout.dim.y-1); x < x1; x++)
      {
	tiles[pos]	    = &grid[y][x];
	tiles[pos++]->loc.x = x*HEXAGONW + (layout.dim.y-1 - y)*HEXAGONW/2;
	Surround(x, y);
      }
    }
    pan_wid = (layout.dim.x - 1 + layout.dim.y - 1) * HEXAGONW;
  }
  else
  {
    for (int y = 0; y < rows_allocated; y++)
    {
      for (int x = 0; x < totalcols; x++)
      {
	grid[y][x].loc.y    = y*HEXAGONR;
	grid[y][x].surround = new Tile*[6];
      }
    }
  }
}


/* ------------------------------   Surround   ------------------------------

  Determine the surrounding hexagons, where the order is:

   5 3	  [ 5 3   ]
  4 - 1   [ 4 - 1 ]    The - is (x, y)
   2 0	  [   2 0 ]

  The reason for this is to help minimise opening hexagons in the cross game.

*/

void HexagonField::Validate(int c, int r)
{
  int w = layout.dim.x - 1,
      h = layout.dim.y - 1;

  if (r < 0 || r > h*2 || c < 0 || c > w+h) return;
  if (r <= h)
  {
    if (c > w + r) return;
  }
  else if (c < r - h) return;

  tile->surround[tile->count++] = &grid[r][c];
}

void HexagonField::Surround(int x, int y)
{
  tile = &grid[y][x];
  tile->count = 0;
  Validate(x+1, y+1);
  Validate(x+1, y);
  Validate(x,	y+1);
  Validate(x,	y-1);
  Validate(x-1, y);
  Validate(x-1, y-1);
}


/* ----------------------------   DisplayTile	-----------------------------

  Display the hexagon t using which bitmap.

*/

void HexagonField::DisplayTile(Tile* t, int which, bool refresh)
{
  Board->DisplayTile(t->loc.x, t->loc.y, HEXAGONW-1, HEXAGONH-1,
		     Hexagon[which], refresh);
}


/* -------------------------   GetBoardDimensions   -------------------------

  Given the dimensions of the minefield, return the dimensions of the board.

*/

Coord HexagonField::GetBoardDimensions(cCoord& dim)
{
  return Coord((dim.x + dim.y - 1) * HEXAGONW,
	       (dim.y * 2 - 1) * HEXAGONR + HEXAGOND);
}


/* ------------------------------   Outline   -------------------------------

  Draw the outline around the minefield. Coyote used six icons to do this, but
  I managed to squeeze them into the two unused numbers.

*/

void HexagonField::Outline(ViewBuffer* buf, int x1, int y1, int x2, int y2)
{
  int dx = (layout.dim.y-1) * (HEXAGONW/2);
  int lx = x1+dx, rx = x2-dx - HEXAGONW/2,
      ty = y1,	  by = y2 - HEXAGONR;

  for (int y = 0; y < layout.dim.y; ++y)
  {
    // Top-left
    BitBltTransparent(buf, lx, ty, Hexagon[7], HEXAGONW/2-1, 0,
					       HEXAGONW-1,   HEXAGONR);
    // Top-right
    BitBltTransparent(buf, rx, ty, Hexagon[7], 0, 0, HEXAGONW/2, HEXAGONR);
    // Bottom-left
    BitBltTransparent(buf, lx, by, Hexagon[8], HEXAGONW/2-1, 0,
					       HEXAGONW-1,   HEXAGONR);
    // Bottom-right
    BitBltTransparent(buf, rx, by, Hexagon[8], 0, 0, HEXAGONW/2, HEXAGONR);

    lx -= HEXAGONW/2; ty += HEXAGONR;
    rx += HEXAGONW/2; by -= HEXAGONR;
  }

  lx = x1+1+dx + HEXAGONW/2;
  ty = y1; by = y2 - HEXAGOND;
  for (int x = 1; x < layout.dim.x; ++x)
  {
    // Top
    BitBltTransparent(buf, lx, ty, Hexagon[7], 0, 0, HEXAGONW-1, HEXAGOND);
    // Bottom
    BitBltTransparent(buf, lx, by, Hexagon[8], 0, HEXAGONR-HEXAGOND,
					       HEXAGONW-1, HEXAGONR);
    lx += HEXAGONW;
  }
}


/* ------------------------------   Validate   ------------------------------

  Ensure the layout does not exceed the limits.

*/

void HexagonField::Validate(Layout& layout)
{
  layout.dim.y = MinHeight >? layout.dim.y <? MaxHeight;
  layout.dim.x = MinWidth  >? layout.dim.x <? GetMaxWidth(layout.dim.y);
  int a        = area(layout.dim);
  layout.mines = (a/10 + (a%10 != 0)) >? layout.mines <? a/2;
}


/* -------------------------   ReadIni & WriteIni   -------------------------

  As you can see, it uses the "[ Hexagonal ]" section and times are kept in
  the "minehex.tym" file.

*/

void HexagonField::ReadIni(IniFile& ini)
{
  Minefield::ReadIni(ini, "Hexagonal", "minehex.tym");
}

void HexagonField::WriteIni(IniFile& ini)
{
  Minefield::WriteIni(ini, "Hexagonal", "minehex.tym");
}
