/*
  Program:      Mine Mayhem
  File:         times.cc
  Date: 	July, 1996 to February,   1997, version 1.00.
		26 October to 3 December,	version 2.00.
		8 December to 4 January,  2001, version 2.10.
		15 January,			version 2.11.

  Everything to do with record times.

  Freeware, Copyright 2001 Jason Hood.

  You are free to use this code, or a portion thereof, as long as an
  appropriate acknowledgement is made.

  v2.00: added Table::str() and Table::Entry::str() functions;
	 added highlight when a new entry is added;
	 added "deviation" option;
	 used "---.---" for "Still waiting" times;
	 modified the level checkbox radiobutton emulation;
	 better centring of the checkboxes.
  v2.10: use my own RadioGroup class, instead of checkboxes;
	 added time, speed and score selection;
	 added clearing of times (saving as text beforehand) and
	  saving best times.
  v2.11: got highlight working again.
*/

#include <time.h>
#include <stdio.h>
#include <tws/message.h>
#include "times.h"
#include "minefld.h"

extern BYTE HighLightColour;


enum
{
  ID_SCORETIME = 30001,
  ID_SCORESPEED,
  ID_SCORESCORE
};


double Layout::speed(double time) const
{
  return mines / time;
}

double Layout::score(double time) const
{
  return sqrt(speed(time) / diff()) * 100;	// speed * density
}


static double deviation_pc(double value, double avg)
{
  return (value - avg) / avg * 100;
}


static Layout get_layout(int level)
{
  Layout lo;

  switch (level)
  {
    case BEGINNER:     lo = field->Beginner;	 break;
    case INTERMEDIATE: lo = field->Intermediate; break;
    case EXPERT:       lo = field->Expert;	 break;
    default:	       lo = field->Custom;	 break;
  }

  return lo;
}


void Table::Entry::clear()
{
  start  = 0;
  finish = 0;
  wrap	 = false;
  strcpy(name, "Still waiting");
  strcpy(date, "??-???-????");
}


// Convert the entry to a string.
// If clear is true, format for the clear game, otherwise the cross.
char* Table::Entry::str(int lev, int score, bool clear, bool deviation,
                        double avg) const 
{
  static char buffer[80],
	      *strstart[]  = { "None", "One ", "Auto"         },
	      *strfinish[] = { "Both", "Open", "Mark", "Itha" },
	      *strwrap[]   = { "",     "Wrap"                 };

  double t = (double)time / 1000;
  avg /= 1000;

  if (*date == '?')
  {
    sprintf(buffer, "---.---  %-20s  %s", name, date);
  }
  else
  {
    if (clear)
    {
      if (score != ID_SCORETIME)
      {
	Layout lo = get_layout(lev);
	if (score == ID_SCORESPEED)
	{
	  t = (deviation) ? deviation_pc(lo.speed(t), lo.speed(avg))
			  : lo.speed(t);
	}
	else // (score == ID_SCORESCORE)
	{
	  t = (deviation) ? deviation_pc(lo.score(t), lo.score(avg))
			  : lo.score(t);
	}
      }
      else if (deviation) t = deviation_pc(t, avg);
      sprintf(buffer, "%7.3f  %-20s  %s  %s  %s  %s",
		      t, name, date,
		      strstart[start], strfinish[finish], strwrap[wrap]);
    }
    else
    {
      if (deviation) t = deviation_pc(t, avg);
      sprintf(buffer, "%7.3f  %-20s  %s  %4d  %4d",
		      t, name, date, spaces, mines);
    }
  }

  return buffer;
}


void Table::save(FILE* file, const char* title, int game) const
{
  if (file != NULL)
  {
    if (title && *title) fprintf(file, "%s\n\n", title);
    int clear = (game == CLEAR);
    unsigned j;
    for (j = 0; j < Entries; ++j)
    {
      if (fastest[j].date[0] == '?') break;
      fprintf(file, "%s\n", fastest[j].str(0, ID_SCORETIME, clear, false, 0));
    }
    if (games <= Entries) j = 0;
    else
    {
      fputc('\n', file);
      if (games <= Entries*2) j = games - Entries;
    }
    for (--j; (int)j >= 0; --j)
    {
      fprintf(file, "%s\n", slowest[j].str(0, ID_SCORETIME, clear, false, 0));
    }
    if (games > 1)
      fprintf(file, "\n%s\n", str(0, ID_SCORETIME, clear));
    else if (games == 0)
      fputs("None played.\n", file);
    fputs("\n\n", file);
  }
}


void Table::best(FILE* file, const char* title, int game) const
{
  if (file != NULL && games > 0)
  {
    if (title && *title) fprintf(file, "%s\n\n", title);
    int clear = (game == CLEAR);
    fprintf(file, "%s\n", fastest[0].str(0, ID_SCORETIME, clear, false, 0));
    if (games > 1)
      fprintf(file, "\n%s\n", str(0, ID_SCORETIME, clear));
    fputs("\n\n", file);
  }
}


// Constructor for an entire times table.
void Table::clear()
{
  games = timetotal = spacetotal = minetotal = 0;

  for (unsigned j = 0; j < Entries; j++)
  {
    fastest[j].clear();
    fastest[j].time   = 999999; 	  // ie. 999.999
    fastest[j].spaces =
    fastest[j].mines  = 0;

    slowest[j].clear();
    slowest[j].time   = 0;
    slowest[j].spaces =
    slowest[j].mines  = 999;
  }
}


/* -------------------------------   Enter   --------------------------------

  Update the number of games and total time for this table. If time is a
  record, then ask for the name and insert it at the appropriate position.
  Return -1 if the time was not a record, otherwise 0 - 4 for the fast times,
  and 5 - 9 for the slow.

*/

int Table::Enter(unsigned time, unsigned spaces, unsigned mines)
{
  int isRecord = -1;

  ++games;
  timetotal  += time;
  spacetotal += spaces;
  minetotal  += mines;

  for (unsigned j = 0; j < Entries; j++)
  {
    if (time <= fastest[j].time)        // Same time is regarded as better
    {                                   //  to reflect the later date
      shift(fastest, j);
      Stats(fastest, j, time, spaces, mines);
      isRecord = j;
      break;
    }
  }
  for (unsigned j = 0; j < Entries; j++)
  {
    if (time >= slowest[j].time)        // Same time is regarded as worse
    {                                   //  to reflect the later date
      shift(slowest, j);
      if (isRecord != -1) slowest[j] = fastest[isRecord];
      else
      {
	Stats(slowest, j, time, spaces, mines);
	isRecord = j + Entries;
      }
      break;
    }
  }
  return isRecord;
}

// Make room for the new entry.
void Table::shift(Entry which[], unsigned from)
{
  // which[from+1..Entries-1] = which[from..Entries-2]
  for (unsigned j = Entries-1; j > from; j--)
  {
    which[j] = which[j-1];
  }
}

// Place the values in the entry.
void Table::Stats(Entry which[], unsigned position,
		  unsigned time, unsigned spaces, unsigned mines)
{
  GetDate(which[position]);
  AskName(which, position);
  which[position].time	 = time;
  which[position].start  = Start;
  which[position].finish = Finish;
  which[position].wrap	 = (isWrap && field->canWrap());
  which[position].spaces = spaces;
  which[position].mines  = mines;
}

// Place the date in the entry as "dd-Mmm-yyyy" (eg. 15-Feb-1997).
void Table::GetDate(Entry& entry)
{
  time_t now;
  time(&now);
  strftime(entry.date, DateLen, "%e-%b-%Y", localtime(&now));
}


// Show the player's position in which table and ask for the name.
void Table::AskName(Entry which[], unsigned position)
{
  static char *place[] = { "", " second", " third", " fourth", " fifth" },
              *fast = "Wow! That is the%s fastest time!",
              *slow = "Unfortunately, that is the%s slowest time.";

  char buffer[80], *message;

  message = (which == fastest) ? fast : slow;
  sprintf(buffer, message, place[position]);
  strcpy(which[position].name, EnterNameDlg(buffer).Run());
}


// Format the statistics as a string.
char* Table::str(int lev, int score, bool clear) const
{
  static char buffer[80];

  if (games == 0) *buffer = 0;
  else
  {
    double avg = ((double)timetotal / games) / 1000;
    if (clear)
    {
      if (score != ID_SCORETIME)
      {
	Layout lo = get_layout(lev);
	if (score == ID_SCORESPEED) avg = lo.speed(avg);
	else /* (score == ID_SCORESCORE) */ avg = lo.score(avg);
      }
      sprintf(buffer, "%7.3f  Average               %d cleared", avg, games);
    }
    else
    {
      char temp[13];
      sprintf(temp, "%d crossed", games);
      sprintf(buffer, "%7.3f  Average               %-11s  %4d  %4d",
		      avg, temp, spacetotal/games, minetotal/games);
    }
  }
  return buffer;
}


/* -------------------------------   Times   -------------------------------- */

Times::Times()
{
  head = prevCustom = new CustomTimes;
  head->next = NULL;
}


Times::~Times()
{
  CustomTimes* list;
  while (head != NULL)
  {
    list = head;
    head = head->next;
    delete list;
  }
}


/* -----------------------------   FindTable   ------------------------------

  Given the level, type of game and automatic options, return a pointer to
  the appropriate table. If level is CUSTOM and there is no matching layout,
  create a new one.

*/

Table* Times::FindTable(int level, int game, bool automark, bool autoopen)
{
  Table* table = NULL;

  switch (level)
  {
    case BEGINNER:     table = Beginner;     break;
    case INTERMEDIATE: table = Intermediate; break;
    case EXPERT:       table = Expert;	     break;
    default: if (field->Custom == prevCustom->layout) table = prevCustom->times;
             else
             {
               CustomTimes* list = head;
               for (; list->next; list = list->next)
               {
		 if (list->layout == field->Custom)
                 {
		   table      = list->times;
                   prevCustom = list;
                   break;
                 }
               }
	       if (table == NULL)
               {
		 list->layout = field->Custom;
		 table	      = list->times;
		 prevCustom   = list;
		 list	      =
		 list->next   = new CustomTimes;
		 list->next   = NULL;
		 field->CustomPopup.Insert(field->Custom);
               }
             }
             break;
  }
  //	       Clear   Cross
  // Normal:	 0	 3
  // AutoMark:	 2	 5
  // AutoOpen:	 1	 4
  return &table[(game == CROSS)*3 + automark*2 + autoopen];
}


/* -------------------------------   Check   --------------------------------

  Determine if time is a record for level, provided it's not a random game,
  or it's not automatic opening AND marking. If it is a record, display it and
  return true.

*/

bool Times::Check(int level, unsigned time, unsigned spaces, unsigned mines)
{
  if (level < BEGINNER || (isAutoMark && isAutoOpen)) return false;

  Table* table	  = FindTable(level);
  int	 isRecord = table->Enter(time, spaces, mines);
  if (isRecord != -1) TimesWindow(*table, level, isRecord).Run();
  return (isRecord != -1);
}


/* ------------------------------   Display   -------------------------------

  Display the times for level. If both automatics are on, turn them off, since
  there is no time for that option. If level is random, use custom if there
  are any custom games stored, otherwise use beginner.

*/

void Times::Display(int level)
{
  bool mark  = isAutoMark,
       open  = isAutoOpen;

  if (level < BEGINNER)
    level = (field->CustomPopup.number == 0) ? BEGINNER : CUSTOM;
  Table* table = FindTable(level);
  if (isAutoMark && isAutoOpen) isAutoMark = isAutoOpen = false;
  TimesWindow(*table, level).Run();
  isAutoMark = mark;
  isAutoOpen = open;
}


/* --------------------------------   Save   --------------------------------

  Write all the times to disk.
  Returns false if the file could not be created, true otherwise.

*/

bool Times::Save(const char* filename)
{
  FILE* file = fopen(filename, "wb");
  if (!file) return false;

  fwrite(Beginner,     sizeof(Beginner),     1, file);
  fwrite(Intermediate, sizeof(Intermediate), 1, file);
  fwrite(Expert,       sizeof(Expert),	     1, file);

  for (CustomTimes* list = head; list->next; list = list->next)
  {
    fwrite(&list->layout, sizeof(list->layout), 1, file);
    fwrite(list->times,   sizeof(list->times),	1, file);
  }

  fclose(file);
  return true;
}


/* --------------------------------   Load   --------------------------------

  Read all the times from disk and create the custom games list and popup menu.
  Returns false if the file could not be opened, true otherwise.

*/

bool Times::Load(const char* filename, CustomPopupRes& custom)
{
  FILE* file = fopen(filename, "rb");
  if (!file) return false;

  fread(Beginner,     sizeof(Beginner),     1, file);
  fread(Intermediate, sizeof(Intermediate), 1, file);
  fread(Expert,       sizeof(Expert),	    1, file);

  CustomTimes* list = head;
  for (;;)
  {
    fread(&list->layout, sizeof(list->layout), 1, file);
    if (feof(file)) break;
    fread(list->times,	 sizeof(list->times),  1, file);
    custom.Insert(list->layout);
    list = list->next = new CustomTimes;
  }
  list->next = NULL;

  fclose(file);
  return true;
}


/* -------------------------------   Delete   --------------------------------

  If the layout has been saved, prompt for confirmation before deleting.
  Returns true if the layout was deleted.

*/

bool Times::Delete(cLayout& lo)
{
  CustomTimes *list, *prev;

  for (prev = list = head; list->next; list = list->next)
  {
    if (list->layout == lo) break;
    prev = list;
  }
  if (list->next == NULL) return false;

  char buf[80];
  sprintf(buf, "Are you sure you want to delete %d x %d, %d mines?",
	       lo.dim.x, lo.dim.y, lo.mines);
  if (MessageBox(buf, "Confirm", MB_YES | MB_NO).Run() == MB_NO) return false;

  if (list == head)
    head = list->next;
  else
    prev->next = list->next;
  delete list;

  return true;
}


/* ----------------------------   TimesWindow	----------------------------- */

static bool ignore_first;		// Ignore initial level event

static char* level_str[] = { "Beginner", "Intermediate", "Expert", "Custom" };
static char* game_str[] = { "Clear", "Clear (auto-open)", "Clear (auto-mark)",
			    "Cross", "Cross (auto-open)", "Cross (auto-mark)" };

static FILE* OpenTimes(const char*);
static void  CloseTimes(FILE*);


enum
{
  ID_DEVIATION = 30004,

  ID_AUTOMARK,
  ID_AUTOOPEN,

  ID_CLEAR,
  ID_CLEARLEVEL,
  ID_CLEARGAME,
  ID_CLEARALL,

  ID_SAVEBEST
};


DEFINE_RESPONSE_TABLE(TimesWindow, TabWin)
  E_BUTTONUP(ID_OK, cmOK)

  E_RADIO_GROUP( ID_SCORETIME, ID_SCORESCORE, RadioScore)
  E_BOXCHECKED(  ID_DEVIATION, cmTable)
  E_BOXUNCHECKED(ID_DEVIATION, cmTable)

  E_BOXCHECKED(  ID_AUTOMARK,  CheckAutoMark)
  E_BOXCHECKED(  ID_AUTOOPEN,  CheckAutoOpen)
  E_BOXUNCHECKED(ID_AUTOMARK,  cmTable)
  E_BOXUNCHECKED(ID_AUTOOPEN,  cmTable)

  E_RADIO_GROUP( CLEAR, CROSS, RadioGame)

  E_RADIO_GROUP( BEGINNER, CUSTOM, RadioLevel)

  E_BUTTONUP(ID_CLEAR,	    cmClear)
  E_BUTTONUP(ID_CLEARLEVEL, cmClearLevel)
  E_BUTTONUP(ID_CLEARGAME,  cmClearGame)
  E_BUTTONUP(ID_CLEARALL,   cmClearAll)

  E_BUTTONUP(ID_SAVEBEST, cmSaveBest)
END_RESPONSE_TABLE


#define TIMES_WIDTH  (64*FixedFont->width)
#define TIMES_HEIGHT (28*FixedFont->height + FixedFont->width)

// Create the times' display and display table.
TimesWindow::TimesWindow(const Table& table, int lev, int highlight) :
  TabWin(NULL, NULL, (ws.GetDeskWidth()  - TIMES_WIDTH)  / 2,
		     (ws.GetDeskHeight() - TIMES_HEIGHT) / 2,
		     (ws.GetDeskWidth()  + TIMES_WIDTH)  / 2 - 1,
		     (ws.GetDeskHeight() + TIMES_HEIGHT) / 2 - 1,
         WA_VISABLE | WA_BORDER | WA_SAVEAREA),
  CustomSave(field->Custom)
{
  const int fw = FixedFont->width,
            fh = FixedFont->height;

  SysFont = FixedFont;
  SysFontBold = FixedFontBold;

  int oyf = FWIDTH + 2*fh + fh/2,
      oys = oyf + Entries*(fh+2) + 4*fh,
      wid = 60*fw;
  for (unsigned j = 0; j < Entries; j++)
  {
    fast[j] = new StaticText(this, "", 2*fw, oyf, wid);
    oyf += fh+2;
    slow[j] = new StaticText(this, "", 2*fw, oys, wid);
    oys += fh+2;
  }
  status = new StaticText(this, "", 2*fw, oyf + fh + fh/2, wid);

  wid = 13*fw;
  int hyt = 3 + 3*(fh+1),
      ox = fw,
      oy = oys + 3*fh, oy1 = oy;

  scores = new RadioGroup(this, ox, oy, ox+wid-1, oy+hyt-1);
  scores->AddButton("&Time",  ID_SCORETIME);
  scores->AddButton("S&peed", ID_SCORESPEED);
  scores->AddButton("&Score", ID_SCORESCORE);
  scores->SelectButton(ID_SCORETIME);

  ox += 3; oy += hyt;
  deviation = new CheckBox(this, ID_DEVIATION, "&Deviation", ox, oy, wid-20);

  hyt -= fh+1;
  ox += -3 + wid + 2*fw; oy = oy1;
  game = new RadioGroup(this, ox, oy, ox+wid-1, oy+hyt-1);
  game->AddButton("C&lear", CLEAR);
  game->AddButton("C&ross", CROSS);
  game->SelectButton(Game);

  ox += 3; oy += hyt;
  automark = new CheckBox(this,ID_AUTOMARK,"Auto&Mark",ox,oy,wid-20,isAutoMark);
  oy += fh+1;
  autoopen = new CheckBox(this,ID_AUTOOPEN,"Auto&Open",ox,oy,wid-20,isAutoOpen);

  ox += -3 + wid + 2*fw; oy = oy1;
  wid = 15*fw; hyt += 2*(fh+1) + 3;
  level = new RadioGroup(this, ox, oy, ox+wid-1, oy+hyt-1, RADIO_TYPE_DIP);
  level->AddButton("&Beginner",     BEGINNER);
  level->AddButton("&Intermediate", INTERMEDIATE);
  level->AddButton("&Expert",       EXPERT);
  // If there are no custom games, don't display anything, since there's no
  // disabled option. (If the level is custom or random, then a new custom
  // game would have been created.)
  if (field->CustomPopup.number != 0)
    level->AddButton("&Custom", CUSTOM);
  ignore_first = true;
  level->SelectButton((lev < BEGINNER) ? CUSTOM : lev);

  wid = 13*fw; hyt = fh+4;
  ox = TIMES_WIDTH - fw - wid;
  clear = new TextButton(this, ID_CLEAR, ox, oy, ox+wid-1, oy+hyt-1, "Clear");
  oy += hyt + -2;
  clearlevel = new TextButton(this, ID_CLEARLEVEL, ox, oy, ox+wid-1, oy+hyt-1,
				    "Clear Level");
  oy += hyt + -2;
  cleargame = new TextButton(this, ID_CLEARGAME, ox, oy, ox+wid-1, oy+hyt-1,
				   "Clear Game");
  oy += hyt + -2;
  clearall = new TextButton(this, ID_CLEARALL, ox, oy, ox+wid-1, oy+hyt-1,
				   "Clear All");

  oy += hyt + 2;
  save = new TextButton(this, ID_SAVEBEST, ox, oy, ox+wid-1, oy+hyt-1,
				   "Save Best");

  wid = StringWidth("OK", FixedFont)*2; hyt = FixedFontBold->height+4;
  ox = (TIMES_WIDTH-wid)/2; oy = TIMES_HEIGHT - fw - hyt;
  ok = new TextButton(this, ID_OK, ox, oy, ox+wid-1, oy+hyt-1, "O&K");

  AddTabStop(scores);
  AddTabStop(game);
  AddTabStop(automark);
  AddTabStop(autoopen);
  AddTabStop(level);
  AddTabStop(ok);

  Paint(); RefreshWindow();
  Print(table, highlight);

  SysFont = SysFont1;
  SysFontBold = SysFontBold1;
}


TimesWindow::~TimesWindow()
{
  for (unsigned j = 0; j < Entries; j++)
  {
    delete fast[j]; delete slow[j];
  }
  delete status;
  delete scores; delete game; delete level;
  delete automark;
  delete autoopen;
  delete deviation;
  delete clear; delete clearlevel; delete cleargame; delete clearall;
  delete save;
  delete ok;
  field->Custom = CustomSave;
}


/* ----------------------------   PaintWindow	-----------------------------

  Draw the frames and display the titles for the fastest and slowest times,
  and draw the boxes around the selections.

*/

void TimesWindow::PaintWindow(int x1, int y1, int x2, int y2)
{
  Window::PaintWindow(x1, y1, x2, y2);

  ViewBuffer* buf = GetSubBuffer(x1, y1, x2, y2);
  const int   fw  = FixedFont->width,
	      fh  = FixedFont->height;

  int oy = FWIDTH + fh/2,
      hyt = 2*fh + 5*(fh+2) + fh/2;
  Frame(buf, fw, oy, TIMES_WIDTH-fw-1, oy+hyt-1);
  oy += hyt + 3*fh;
  InvFrame(buf, fw, oy, TIMES_WIDTH-fw-1, oy+hyt-1);
  int len = StringWidth("FASTEST TIMES", FixedFont);
  WriteText(buf, (TIMES_WIDTH-len)/2, FWIDTH+fh,
                 HighLightColour, FixedFont, "FASTEST TIMES");
  len = StringWidth("SLOWEST TIMES", FixedFont);
  WriteText(buf, (TIMES_WIDTH-len)/2, oy+hyt-fh-fh/2,
                 HighLightColour, FixedFont, "SLOWEST TIMES");

  oy += hyt + fh;
  hyt = 3 + 4*(fh+1) + 3;
  int wid = 13*fw,
      ox = fw;
  InvFrame(buf, ox, oy, ox+wid-1, oy+hyt-1);
     Frame(buf, ox+1, oy+1, ox+wid-2, oy+hyt-2);

  ox += wid + 2*fw;
  InvFrame(buf, ox, oy, ox+wid-1, oy+hyt-1);
     Frame(buf, ox+1, oy+1, ox+wid-2, oy+hyt-2);

  DeleteBuffer(buf);
}


/* -------------------------------   Print   --------------------------------

  Update the window with the values from table. If highlight is not -1, display
  the appropriate time using the highlight colour.

*/

void TimesWindow::Print(const Table& table, int highlight)
{
  int	 lev = level->GetSelectedButton();
  int  score = scores->GetSelectedButton();
  bool clear = (game->GetSelectedButton() == CLEAR);
  bool	 dev = deviation->isChecked();
  double avg = (table.games) ? (double)table.timetotal / table.games : 0;

  for (unsigned j = 0; j < Entries; j++)
  {
    fast[j]->SetText(table.fastest[j].str(lev, score, clear, dev, avg));
    slow[j]->SetText(table.slowest[Entries-1-j].str(lev, score, clear,dev,avg));
  }
  status->SetText(table.str(lev, score, clear));

  if (highlight >= 0 && highlight < (int)Entries*2)
  {
    // 010115: Something I've done between 2.00 and 2.10 requires this.
    ws.CheckForEvents();
    StaticText* which = (highlight < (int)Entries) ? fast[highlight] :
				       slow[Entries-1 - (highlight - Entries)];
    WriteText(which->GetBuffer(), 0, 0, HighLightColour, FixedFont,
	      which->GetText());
    which->RefreshWindow();
  }
}


void TimesWindow::Run()
{
  SysFont = FixedFont;
  SysFontBold = FixedFontBold;
  SetFocus(ok);
  CatchMouse();
  CatchKeys();
  ws.RunEvents();
  ws.ResetEvents();
  ReleaseMouse();
  ReleaseKeys();
  SysFont = SysFont1;
  SysFontBold = SysFontBold1;
}


// Cross games do not have speed or score, so make sure time is selected.
void TimesWindow::RadioGame(int ID)
{
  if (ID == CROSS && scores->GetSelectedButton() != ID_SCORETIME)
    scores->SelectButton(ID_SCORETIME);
  else
    cmTable();
}

void TimesWindow::RadioScore(int ID)
{
  if (game->GetSelectedButton() == CROSS && ID != ID_SCORETIME)
    scores->SelectButton(ID_SCORETIME);
  else
    cmTable();
}


// Since there's no automatic opening AND marking, uncheck the other one if
// it's checked.
void TimesWindow::CheckAutoMark()
{
  if (autoopen->isChecked()) autoopen->UnCheck();
  cmTable();
}

void TimesWindow::CheckAutoOpen()
{
  if (automark->isChecked()) automark->UnCheck();
  cmTable();
}


// For custom games, pop up a menu with the stored layouts. If the menu is
// cancelled, restore the previous selection.
void TimesWindow::RadioLevel(int ID)
{
  static int old_ID;

  if (ignore_first)			// Ignore the first call, from when
  {					//  the window is created.
    ignore_first = false;
    old_ID = ID;
    return;
  }

  if (ID == CUSTOM)
  {
    const int fw = FixedFont->width,
	      fh = FixedFont->height;

    int ox = absx + fw + 13*fw + 2*fw + 13*fw + 2*fw + 3 + 20,
	oy = absy + FWIDTH + 2*fh + fh/2 + 2*Entries*(fh+2) + 4*fh + 3*fh
	     + 3 + 4*(fh+1) + 1;
    if (CustomPopupMenu(NULL, field->CustomPopup, ox, oy).Run())
    {
      cmTable();
      old_ID = CUSTOM;
    }
    else if (old_ID != CUSTOM)
      level->SelectButton(old_ID);

    CatchKeys();			// It seems PopupMenu releases keys
  }
  else
  {
    old_ID = ID;
    cmTable();
  }
}


/* ------------------------------   CmTable   -------------------------------

  Find the table corresponding to the selections and display it.

*/

void TimesWindow::cmTable(int)
{
  Table* table = field->times.FindTable(level->GetSelectedButton(),
					game->GetSelectedButton(),
					automark->isChecked(),
					autoopen->isChecked());
  Print(*table);
}


/* ------------------------------   CmClear   -------------------------------

  Clear the times currently displayed.

*/

void TimesWindow::cmClear()
{
  if (MessageBox("Are you sure you want to clear these times?", "Confirm",
		 MB_YES | MB_NO).Run() == MB_NO) return;

  int lev = level->GetSelectedButton(),
      gopt = game->GetSelectedButton(),
      amark = automark->isChecked(),
      aopen = autoopen->isChecked();

  int num = (gopt == CROSS)*3 + aopen + (amark << 1);
  Table* table = field->times.FindTable(lev, gopt, amark, aopen);
  char buf[80];
  if (lev == CUSTOM)
  {
    sprintf(buf, "%s (%d x %d, %d mines) %s", level_str[3],
					      field->Custom.dim.x,
					      field->Custom.dim.y,
					      field->Custom.mines,
					      game_str[num]);
  }
  else sprintf(buf, "%s %s", level_str[lev-BEGINNER], game_str[num]);
  FILE* file = OpenTimes(buf);
  table->save(file, NULL, gopt);
  table->clear();
  CloseTimes(file);
  Print(*table);
}


/* ----------------------------   CmClearLevel   ----------------------------

  Clear all the times associated with the current level.

*/

void TimesWindow::cmClearLevel()
{
  int lev = level->GetSelectedButton();
  char buf[80];
  sprintf(buf, "Are you sure you want to clear all the %s times?",
	       level_str[lev - BEGINNER]);
  if (MessageBox(buf, "Confirm", MB_YES | MB_NO).Run() == MB_NO) return;

  Table* table = field->times.FindTable(lev, 0, 0, 0);
  if (lev == CUSTOM)
  {
    sprintf(buf, "%s (%d x %d, %d mines)", level_str[3],
					   field->Custom.dim.x,
					   field->Custom.dim.y,
					   field->Custom.mines);
  }
  else strcpy(buf, level_str[lev - BEGINNER]);
  FILE* file = OpenTimes(buf);
  int gopt = CLEAR;
  for (int j = 0; j < 6; ++j)
  {
    if (j == 3) gopt = CROSS;
    table[j].save(file, game_str[j], gopt);
    table[j].clear();
  }
  CloseTimes(file);
  Print(*table);
}


/* ----------------------------   CmClearGame   -----------------------------

  Clear all the times associated with the current game type.

*/

void TimesWindow::cmClearGame()
{
  int gopt = game->GetSelectedButton();
  int num = (gopt == CROSS)*3 + autoopen->isChecked()
			      + (automark->isChecked() << 1);
  char buf[80];
  sprintf(buf, "Are you sure you want to clear all the %s times?",
	       game_str[num]);
  if (MessageBox(buf, "Confirm", MB_YES | MB_NO).Run() == MB_NO) return;

  FILE* file = OpenTimes(game_str[num]);

  field->times.Beginner[num].save(file, level_str[0], gopt);
  field->times.Intermediate[num].save(file, level_str[1], gopt);
  field->times.Expert[num].save(file, level_str[2], gopt);

  field->times.Beginner[num].clear();
  field->times.Intermediate[num].clear();
  field->times.Expert[num].clear();

  for (Times::CustomTimes* list = field->times.head; list->next;
       list = list->next)
  {
    sprintf(buf, "%s (%d x %d, %d mines)", level_str[3],
					   list->layout.dim.x,
					   list->layout.dim.y,
					   list->layout.mines);
    list->times[num].save(file, buf, gopt);
    list->times[num].clear();
  }

  CloseTimes(file);

  Print(field->times.Beginner[num]);
}


/* -----------------------------   CmClearAll   -----------------------------

  Clear all times of the field.

*/

static void clear_level(FILE* file, Table* table, const char* title)
{
  if (file != NULL)
  {
    int k = fprintf(file, "%s\n", title);
    for (; k > 1; --k) fputc('-', file);
    fputs("\n\n", file);

    int gopt = CLEAR;
    for (int j = 0; j < 6; ++j)
    {
      if (j == 3) gopt = CROSS;
      table[j].save(file, game_str[j], gopt);
      table[j].clear();
    }
  }
}


void TimesWindow::cmClearAll()
{
  if (MessageBox("Are you sure you want to clear ALL times?", "Confirm",
		 MB_YES | MB_NO).Run() == MB_NO) return;

  FILE* file = OpenTimes("");

  clear_level(file, field->times.Beginner,     level_str[0]);
  clear_level(file, field->times.Intermediate, level_str[1]);
  clear_level(file, field->times.Expert,       level_str[2]);

  for (Times::CustomTimes* list = field->times.head; list->next;
       list = list->next)
  {
    char buf[80];
    sprintf(buf, "%s (%d x %d, %d mines)\n", level_str[3],
					     list->layout.dim.x,
					     list->layout.dim.y,
					     list->layout.mines);
    clear_level(file, list->times, buf);
  }

  CloseTimes(file);
  Print(field->times.Beginner[0]);
}


/* -----------------------------   CmSaveBest	-----------------------------

  Save all the best times of the field.

*/

static void save_best(FILE* file, Table* table, const char* title)
{
  if (file != NULL)
  {
    int k = fprintf(file, "%s\n", title);
    for (; k > 1; --k) fputc('-', file);
    fputs("\n\n", file);

    int gopt = CLEAR;
    for (int j = 0; j < 6; ++j)
    {
      if (j == 3) gopt = CROSS;
      table[j].best(file, game_str[j], gopt);
    }
  }
}


void TimesWindow::cmSaveBest()
{
  FILE* file = OpenTimes(NULL);

  save_best(file, field->times.Beginner,     level_str[0]);
  save_best(file, field->times.Intermediate, level_str[1]);
  save_best(file, field->times.Expert,	     level_str[2]);

  for (Times::CustomTimes* list = field->times.head; list->next;
       list = list->next)
  {
    char buf[80];
    sprintf(buf, "%s (%d x %d, %d mines)\n", level_str[3],
					     list->layout.dim.x,
					     list->layout.dim.y,
					     list->layout.mines);
    save_best(file, list->times, buf);
  }

  CloseTimes(file);
}


FILE* OpenTimes(const char* title)
{
  FILE* file = fopen((title == NULL) ? "best.txt" : "times.txt", "a");
  if (file != NULL)
  {
    time_t now;
    char buf[80];
    time(&now);
    strftime(buf, sizeof(buf), "%e %B, %Y", localtime(&now));
    const char* type = (field == squarefield)  ? "Square" :
		       (field == recthexfield) ? "Hexagon" :
						 "Hexagonal";
    if (title == NULL) title = "best";
    int k = fprintf(file, "%s field %s%stimes as at %s\n", type, title,
			  (*title) ? " " : "", buf);
    for (; k > 1; --k) fputc('=', file);
    fputs("\n\n", file);
  }
  return file;
}


void CloseTimes(FILE* file)
{
  if (file != NULL)
  {
    for (int j = 0; j < 78; ++j) putc('-', file);
    fputs("\n\n\n", file);
    fclose(file);
  }
}


/* ----------------------------   EnterNameDlg	 ---------------------------- */

#define NAME_WIDTH  (50*SysFont1->width)
#define NAME_HEIGHT (12*SysFont1->height)

#define NAME_ID 30000

DEFINE_RESPONSE_TABLE(EnterNameDlg, Window)
  E_KEYCODE(0x0d, 0x1c, KeyEnter)
  E_BUTTONUP(ID_OK,	cmOK)
END_RESPONSE_TABLE

char EnterNameDlg::lastname[NameLen] = { 0 };


// Create the enter name dialog box, where message indicates whether it was a
// fast or slow time. The name is entered using the fixed font.
EnterNameDlg::EnterNameDlg(const char* message) :
  Window(NULL, "Congratulations", (ws.GetDeskWidth() -  NAME_WIDTH)  / 2,
				  (ws.GetDeskHeight() - NAME_HEIGHT) / 2,
				  (ws.GetDeskWidth() +	NAME_WIDTH)  / 2 - 1,
				  (ws.GetDeskHeight() + NAME_HEIGHT) / 2 - 1,
         WA_VISABLE | WA_CAPTION | WA_BORDER | WA_SAVEAREA)
{
  int len = StringWidth(message, SysFont);
  msg = new StaticText(this, message,
                       (NAME_WIDTH-len)/2, SysFont->height*3, len);
  len = NameLen*FixedFont->width;
  SysFont = FixedFont;
  name = new EditText(this, NAME_ID, "", NameLen,
                      (NAME_WIDTH-len)/2, SysFont1->height*6, len);
  SysFont = SysFont1;
  len = StringWidth("OK", SysFont)*2;
  int ox = (NAME_WIDTH-len)/2, oy = cy2-SysFont->height*2-4;
  ok = new TextButton(this, ID_OK, ox,oy, ox+len-1, oy+SysFont->height+3, "OK");
  SetFocus(name);
  name->SetText(lastname);
}


EnterNameDlg::~EnterNameDlg()
{
  delete msg;
  delete name;
  delete ok;
}


// Write the prompt onto the window.
void EnterNameDlg::PaintWindow(int x1, int y1, int x2, int y2)
{
  Window::PaintWindow(x1, y1, x2, y2);

  ViewBuffer* buf = GetSubBuffer(x1, y1, x2, y2);
  WriteText(buf, (NAME_WIDTH-NameLen*FixedFont->width)/2, SysFont->height*5,
                 HighLightColour, SysFontItallic, "Please enter your name:");
  DeleteBuffer(buf);
}


const char* EnterNameDlg::Run()
{
  Paint();
  SysFont = FixedFont;
  name->Paint();
  RefreshWindow();
  CatchMouse();
  CatchKeys();
  ws.RunEvents();
  ws.ResetEvents();
  ReleaseMouse();
  ReleaseKeys();
  SysFont = SysFont1;
  strcpy(lastname, name->GetText());
  return (*lastname == 0) ? "Anonymous" : lastname;
}
