/*
  Program:      Mine Mayhem
  File:         times.h
  Date: 	July, 1996 to February,   1997, version 1.00.
		26 October to 3 December,	version 2.00.
		8 December to 4 January,  2001, version 2.10.
		15 January,			version 2.11.

  Definitions for record times.

  Freeware, Copyright 2001 Jason Hood.

  You are free to use this code, or a portion thereof, as long as an
  appropriate acknowledgement is made.

  v2.00: added Table::str() and Table::Entry::str() functions;
	 added highlight parameter.
  v2.10: use my own RadioGroup class, instead of checkboxes;
	 added time, speed and score selection;
	 added ability to clear times (saving as text beforehand)
	  and save all best times.
*/

#ifndef _times_h
#define _times_h

#include <tws/tabwin.h>
#include <tws/static.h>
#include <tws/edit.h>
#include <tws/checkbox.h>
#include <tws/menu.h>
#include <tws/button.h>

#include "layout.h"
#include "options.h"
#include "custom.h"
#include "radiogrp.h"

extern FONT *FixedFont, *FixedFontBold, *SysFont1, *SysFontBold1;
extern int Game;			// mine.cc


const unsigned NameLen = 20 + 1,
               DateLen = 11 + 1,        // "dd-Mmm-yyyy\0"
               Entries = 5;


// This class holds the five fastest and slowest times.
class Table
{
friend class TimesWindow;

public:
  Table() { clear(); }
  void clear();

  int	Enter(unsigned time, unsigned spaces, unsigned mines);
					// Check if time (in ms) is a record
  char* str(int level, int score, bool clear) const;
					// Format the statistics as a string

  void	save(FILE* file, const char* title, int game) const;
  void	best(FILE* file, const char* title, int game) const;

protected:
  unsigned games,                       // The number of games completed
           timetotal,                   // The total time for all games
           spacetotal,                  // Total number of spaces cleared
           minetotal;                   // Total number of mines found
					// (Above two for the cross game)

  struct Entry
  {
    void  clear();
    char* str(int level, int score, bool clear, bool dev, double avg) const;
					// Format the entry as a string

    unsigned time;                      // In milliseconds
    char     name[NameLen],
	     date[DateLen];
    int      start,			// Starting condition
	     finish;			// Finishing condition
    bool     wrap;			// Wrap-around (all three for clear)
    unsigned spaces,			// Spaces cleared
	     mines;			// Mines flagged (both for cross)

  } fastest[Entries], slowest[Entries];

  void shift(Entry which[], unsigned from);	// Make room for the new entry
  void GetDate(Entry& entry);
  void AskName(Entry which[], unsigned position);
  void Stats(Entry which[], unsigned position,
	     unsigned time, unsigned spaces, unsigned mines);
};


// This class holds the tables for each level and option.
class Times
{
friend class TimesWindow;

public:
  Times();
  ~Times();

  bool Check(int level, unsigned time, unsigned spaces, unsigned mines);
                                        // Is time a record for level?
  void Display(int level);              // Display the times for level

  bool Save(const char* filename);
  bool Load(const char* filename, CustomPopupRes& custom);

  bool Delete(cLayout& lo);		// Delete the associated custom layout

protected:
  Table* FindTable(int level, int game = Game, bool automark = isAutoMark,
                                               bool autoopen = isAutoOpen);

  Table Beginner[6], Intermediate[6], Expert[6];

  struct CustomTimes
  {
    Layout	 layout;
    Table	 times[6];
    CustomTimes* next;
  } *head, *prevCustom;
};


// This is the class that does all the displaying.
class TimesWindow : public TabWin
{
public:
  TimesWindow(const Table& table, int level, int highlight = -1);
  ~TimesWindow();

  void Run();

protected:
  void PaintWindow(int x1, int y1, int x2, int y2);

  void Print(const Table& table, int highlight = -1);

  void cmOK() { ws.StopRunningEvents(); }

  void CheckAutoMark();
  void CheckAutoOpen();

  void RadioGame(int);
  void RadioScore(int);
  void RadioLevel(int);

  void cmTable(int = 0);

  void cmClear();
  void cmClearLevel();
  void cmClearGame();
  void cmClearAll();

  void cmSaveBest();

  DECLARE_RESPONSE_TABLE;

  StaticText *fast[Entries],
	     *status,
	     *slow[Entries];
  RadioGroup *scores, *game, *level;
  CheckBox   *automark,
	     *autoopen,
	     *deviation;
  TextButton *clear, *clearlevel, *cleargame, *clearall;
  TextButton *save;

  TextButton *ok;

  Layout CustomSave;
};


// This is the dialog box to ask for the player's name.
class EnterNameDlg : public Window
{
public:
  EnterNameDlg(const char* message);
  ~EnterNameDlg();

  const char* Run();

protected:
  void PaintWindow(int x1, int y1, int x2, int y2);

  void KeyEnter(int) { ws.StopRunningEvents(); }
  void cmOK()	     { ws.StopRunningEvents(); }

  DECLARE_RESPONSE_TABLE;

  StaticText* msg;
  EditText*   name;
  TextButton* ok;

  static char lastname[NameLen];
};

#endif
